(function (exports) {
'use strict';

function wrapPrototype(klass, method, fn){
    var oldMethod = klass.prototype[method];
    var newMethod = fn(oldMethod);

    klass.prototype[method] = newMethod;
}

/*:
 * @plugindesc 視線、聴覚トリガを実現します。
 *
 * @author liply
 *
 * @help
 * トリガを設定したいページで、注釈コマンドに以下のコマンドを記述します。
 *
 * triggerHear
 * 聴覚を実現します。指定された半径の円状に当たり判定を展開します。
 * 半径単位はキャラクタ数です。
 *
 * 例
 * triggerHear 3
 * 半径３のプレイヤーに判定
 *
 * triggerSee
 *
 * 視覚を実現します。指定された半径、角度の扇状に当たり判定を展開します。
 * その際、イベントの向きを考慮します。
 *
 * 例
 * triggerSee 3 90
 * 半径３、角度９０度のプレイヤーに判定
 *
 * triggerSquare
 * 視界長方形上に当たり判定を展開します。
 *
 * 例
 * triggerSquare 3 10
 * 幅３、長さ１０のプレイヤーに判定
 *
 */

var ISQ2 = 1 / Math.sqrt(2);
var TILE_SIZE = 48;
var AREA_COLOR = 0xff0000;
var AREA_ALPHA = 0.3;

function toXy(d){
    switch(d){
        case 1:
            return {x: -ISQ2, y: ISQ2};
        case 2:
            return {x: 0, y: 1};
        case 3:
            return {x: ISQ2, y: ISQ2};
        case 4:
            return {x:-1, y: 0};
        case 5:
            return {x: 0,y: 0};
        case 6:
            return {x: 1, y: 0};
        case 7:
            return {x: -ISQ2, y: -ISQ2};
        case 8:
            return {x: 0, y: -1};
        case 9:
            return {x: ISQ2, y: -ISQ2};
    }

    return {x:0,y:0};
}

function isEventReady(event){
    return !$gameMap.isEventRunning() && !event._erased;
}

wrapPrototype(Game_Event, 'setupPageSettings', function (old){ return function(){
    old.call(this);
    var triggers = readTrigger(this.page());

    delete this._triggerHear;
    delete this._triggerSee;
    delete this._triggerSquare;

    this._triggerHear = triggers.hear;
    this._triggerSee = triggers.see;
    this._triggerSquare = triggers.square;
}; });

wrapPrototype(Game_Event, 'update', function (old){ return function(){
    old.call(this);
    if(this._triggerHear){
        checkTriggerHear(this, this._triggerHear);
    }
    if(this._triggerSee){
        checkTriggerSee(this, this._triggerSee);
    }
    if(this._triggerSquare){
        checkTriggerSquare(this, this._triggerSquare);
    }
}; });

//triggerHear 3
//triggerSee 3 360
//triggerSquare 3 10
function readTrigger(page){
    var trigger = {};
    page.list.forEach(function (cmd){
        if(cmd.code === 108 || cmd.code === 408){
            var p = cmd.parameters[0];
            var splat = p.split(' ');
            var triggerName = splat[0].toLowerCase();
            if(triggerName === 'triggerhear'){
                trigger.hear = {range: +splat[1]};
            }else if(triggerName === 'triggersee'){
                trigger.see = {range: +splat[1], angle: +splat[2]};
            }else if(triggerName === 'triggersquare'){
                trigger.square = {width: +splat[1], length: +splat[2]};
            }
        }
    });

    return trigger;
}

function checkTriggerHear(event, param){
    var dx =  $gamePlayer._realX - event._realX;
    var dy =  $gamePlayer._realY - event._realY;

    var l = Math.sqrt(dx*dx+dy*dy);
    if(l <= param.range && isEventReady(event)){
        event.start();
    }
}

function checkTriggerSee(event, param){
    var dx = $gamePlayer._realX - event._realX;
    var dy = $gamePlayer._realY - event._realY;
    var l = Math.sqrt(dx*dx+dy*dy);

    if(l < param.range && isEventReady(event)){
        var nx = dx / l;
        var ny = dy / l;

        var dir = toXy(event.direction());
        var cos = nx * dir.x + ny * dir.y;
        if(!isNaN(cos)){
            var theta = Math.acos(cos);
            var angleRad = (param.angle * Math.PI / 180) / 2;
            if(theta < angleRad){
                event.start();
            }
        }
    }
}

function dir2square(direction, cx, cy, width, height){
    var left, top, right, bottom;

    switch(direction){
        case 2:
            left = cx - width / 2;
            top = cy;
            right = cx + width / 2;
            bottom = cy + height;
            break;

        case 4:
            left = cx - height;
            top = cy - width / 2;
            right = cx;
            bottom = cy + width / 2;
            break;

        case 6:
            left = cx;
            top = cy - width / 2;
            right = cx + height;
            bottom = cy + width / 2;
            break;

        case 8:
            left = cx - width / 2;
            top = cy - height;
            right = cx + width / 2;
            bottom = cy;
            break;
    }

    return {left: left, top: top, right: right, bottom: bottom};
}

function checkTriggerSquare(event, param){
    var cx = event._realX + 0.5;
    var cy = event._realY + 0.5;

    var square = dir2square(event.direction(),
        cx, cy,
        param.width, param.length);

    var x = $gamePlayer._realX + 0.5;
    var y = $gamePlayer._realY + 0.5;

    if(square.left <= x && x <= square.right &&
        square.top <= y && y <= square.bottom && isEventReady(event)){
        event.start();
    }
}

function dir2angle(dir){
    switch(dir){
        case 2:
            return +Math.PI/2;
        case 4:
            return Math.PI;
        case 6:
            return 0;
        case 8:
            return -Math.PI/2;
    }
}

wrapPrototype(Sprite_Character, 'initialize', function (old){ return function(){
    old.apply(this, arguments);
    this._areaGraphics = new PIXI.Graphics();
    // this._areaGraphics.position.x = TILE_SIZE / 2;
    this._areaGraphics.position.y = -TILE_SIZE / 2;

    this._areaSprite = new Sprite();
    // this._areaSprite.position.x = TILE_SIZE / 2;
    // this._areaSprite.position.y = TILE_SIZE / 2;


    this.addChild(this._areaSprite);
    this._areaSprite.addChild(this._areaGraphics);
}; });

wrapPrototype(Sprite_Character, 'update', function (old){ return function(){
    old.apply(this, arguments);

    if(this._character._erased){
        this._areaGraphics.visible = false;
    }else if(this._character._triggerHear){
        drawTriggerHear(this._areaGraphics, this._character._triggerHear);
    } else if(this._character._triggerSee){
        drawTriggerSee(this, this._areaGraphics, this._character._triggerSee);
    } else if(this._character._triggerSquare){
        drawTriggerSquare(this, this._areaGraphics, this._character._triggerSquare);
    }else{
        this._areaGraphics.visible = false;
    }
}; });

function drawTriggerHear(graphics, param){
    var range = param.range * TILE_SIZE;
    graphics.visible = true;

    if(!param.drawn){
        graphics.clear();
        graphics.beginFill(AREA_COLOR, AREA_ALPHA);
        graphics.drawCircle(0, 0, range);
        graphics.endFill();

        param.drawn = true;
    }
}


function drawTriggerSee(sprite, graphics, param){
    var angle = param.angle * Math.PI / 180;
    var range = param.range * TILE_SIZE;
    graphics.visible = true;

    if(!param.drawn){
        graphics.clear();
        graphics.beginFill(AREA_COLOR, AREA_ALPHA);
        graphics.moveTo(0,0);
        graphics.arc(0, 0, range, -angle/2, angle/2, false);
        graphics.lineTo(0,0);
        graphics.endFill();

        param.drawn = true;
    }

    graphics.rotation = dir2angle(sprite._character.direction());
}


function drawTriggerSquare(sprite, graphics, param){
    var width = param.width * TILE_SIZE;
    var length = param.length * TILE_SIZE;
    graphics.visible = true;

    if(!param.drawn){
        graphics.clear();
        graphics.beginFill(AREA_COLOR, AREA_ALPHA);
        graphics.drawRect(0, -width/2, length, width);
        graphics.endFill();

        param.drawn = true;
    }

    graphics.rotation = dir2angle(sprite._character.direction());
}

exports.toXy = toXy;

}((this.liply = this.liply || {})));